using DrWatson
@quickactivate "StochApprox"
using Revise
using saControlLimits
import saControlLimits: spsaCL
using StatsBase
using ArgParse
using Random, Distributions
using DataFrames
using JLD2
using TimerOutputs

include(srcdir("settings", "settings.jl"))

DrWatson.default_allowed(c) = (Real, String, Symbol, Vector)
DrWatson.savename(S::Settings, suffix::String) = savename(struct2ntuple(S.PROC), suffix)

argTbl = ArgParseSettings()
@add_arg_table argTbl begin
    "--seed", "-s"
        help = "Simulation seed"
end

parsedArgs = parse_args(ARGS, argTbl)
SD = parse(Int64, parsedArgs["seed"])

simsDir = "bisection"
simsDir_nelder = "nelder"
simsDir_evo = "evo"

abstol = 1e-05
NminSPSA = 300
OPTS = OptimizationSettings(
    hStart = 0.001,
    nAdaptiveGain = 20,
    initialStepSize = [0.2],
    expNumberLossEval = NminSPSA,
    NminSPSA = NminSPSA,
    NfixedSPSA = 100,
    gg = 0.05,
    abstol = abstol,
    gavg = 100,
    NfixedSA = 100,
    maxiterSA=100,
    maxiterSPSA = 100000,
    minPar = [0.01],
    maxPar = [0.99],
    comparison = nothing)


time()      #? First run for compilation
# ---- Bisection CUSUM ---- #
println("--- Starting CUSUM ---")
rng = MersenneTwister(SD)
deltas = 0.5:0.25:3.0
ARL0 = 370
nsim_bisection = 10000
minPar = [0.001]
maxPar = [4.0]
for delta in deltas
    seed = abs(rand(rng, Int))
    filename = savename(Dict("delta" => delta, "ARL0" => ARL0, "SD" => SD), "jld2")
    filepath = datadir(simsDir, "CUSUM", filename)
    if isfile(filepath)
        
    else
        #? First run for compilation
        bisectionOpt([0.1], [0.1], runCUSUM, ARL0, shift = delta, m = 3, verbose = false, maxiter=1, nsim = 1, seed = seed)

        t_bisec = time()
        opt_bisection = bisectionOpt(minPar, maxPar, runCUSUM, ARL0, shift = delta, m = 3, verbose = false, nsim = nsim_bisection, seed = seed)
        dt_bisec = time() - t_bisec

        par0 = [1.0]
        t1_spsa = time()
        optH = saCL(0.1, (h,m,d) -> runCUSUM(h,m,d, par=par0), ARL0, verbose=false)
        dt1_spsa = time() - t1_spsa
        OPTS_run = OptimizationSettings(OPTS,
                    hStart = optH[:h],
                    initialStepSize = [0.2],
                    minPar = minPar,
                    maxPar = maxPar,
                    #! Put comparison to compare accuracy
                    comparison = nothing
                    )

        PROC = ProcessSettings(par0 = par0, ARL0 = ARL0, delta=delta, seed=seed)
        settings = Settings(OPTS_run, PROC)

        settings_compile = Settings(settings, OPTIM = OptimizationSettings(OPTS_run, maxiterSPSA=1))
        spsaCL(settings_compile, runCUSUM)


        t_spsa = time()
        opt_spsa = spsaCL(settings, runCUSUM)
        dt_spsa = time() - t_spsa
        dt_spsa += dt1_spsa


        saveobj = Dict("delta" => delta,
                        "ARL0" => ARL0,
                        "parSPSA" => opt_spsa[:opt][1], 
                        "nIterICspsa" => opt_spsa[:nIterIC] + optH[:iter],
                        "nIterOCspsa" => opt_spsa[:nIterOC],
                        "parBisec" => opt_bisection[:opt][1], 
                        "nIterICbisec" => opt_bisection[:nIterIC],
                        "nIterOCbisec" => opt_bisection[:nIterOC],
                        "time_spsa" => dt_spsa,
                        "time_bisec" => dt_bisec,
                        )
        safesave(filepath, saveobj)
    end

    filename = savename(Dict("delta" => delta, "ARL0" => ARL0, "SD" => SD, "type" => "NelderMead"), "jld2")
    filepath = datadir(simsDir_nelder, "CUSUM", filename)
    if isfile(filepath)
        
    else
        #? First run for compilation
        NelderMeadOpt([1.0], minPar, maxPar, runCUSUM, ARL0, shift=delta, maxiter=1, seed=seed)

        t_nelder = time()
        opt_nelder = NelderMeadOpt([1.0], minPar, maxPar, runCUSUM, ARL0, shift=delta, seed=seed)
        dt_nelder = time() - t_nelder

        saveobj = Dict("delta" => delta,
                        "ARL0" => ARL0,
                        "parNelder" => opt_nelder[2],
                        "time_nelder" => dt_nelder,
                        )
        safesave(filepath, saveobj)
    end
end


# ---- Bisection EWMA ---- #
println("--- Starting EWMA ---")
rng = MersenneTwister(SD)
deltas = [0.5, 1.0, 1.5, 2.0]
ARL0 = 370
nsim_bisection = 10000
minPar = fill(0.001, 1)
maxPar = fill(0.999, 1)
ncomparison = 10^6
for delta in deltas
    seed = abs(rand(rng, Int))
    filename = savename(Dict("delta" => delta, "ARL0" => ARL0, "SD" => SD), "jld2")
    filepath = datadir(simsDir, "EWMA", filename)
    if isfile(filepath)
        
    else
        #? First run for compilation
        bisectionOpt(minPar, maxPar, runEWMA, ARL0, shift = delta, maxiter=1, m = 3, verbose = false, nsim = 1, seed = seed, eps3 = abstol)


        t_bisec = time()
        opt_bisection = bisectionOpt(minPar, maxPar, runEWMA, ARL0, shift = delta, m = 3, verbose = false, nsim = nsim_bisection, seed = seed)
        dt_bisec = time() - t_bisec

        par0 = fill(0.2, 1)
        t1_spsa = time()
        optH = saCL(10.0, (h,m,d) -> runEWMA(h,m,d, par=par0), ARL0, verbose=false)
        dt1_spsa = time() - t1_spsa
        OPTS_run = OptimizationSettings(OPTS,
                    hStart = optH[:h],
                    initialStepSize = fill(0.1, 1),
                    minPar = minPar,
                    maxPar = maxPar,
                    comparison = nothing
                    )

        PROC = ProcessSettings(par0 = par0, ARL0 = ARL0, delta=delta, seed=seed)
        settings = Settings(OPTS_run, PROC)

        settings_compile = Settings(settings, OPTIM = OptimizationSettings(OPTS_run, maxiterSPSA=1))
        #? First run for compilation
        spsaCL(settings_compile, runEWMA)

        t_spsa = time()
        opt_spsa = spsaCL(settings, runEWMA)
        dt_spsa = time() - t_spsa
        dt_spsa += dt1_spsa


        h_opt_b = saCL(1.0, (h,m,d) -> runEWMA(h,m,d, par=opt_bisection[:opt]), ARL0)[:h]
        h_opt_spsa = saCL(1.0, (h,m,d) -> runEWMA(h,m,d, par=opt_spsa[:opt]), ARL0)[:h]

        seed = abs(rand(rng, Int))
        OC_rl_bisec = mean([runEWMA(h_opt_b, 10^5, 0.0, par=opt_bisection[:opt], shift=delta, seed = seed+i)[:rl] for i in 1:ncomparison])
        OC_rl_spsa = mean([runEWMA(h_opt_spsa, 10^5, 0.0, par=opt_spsa[:opt], shift=delta, seed = seed+i)[:rl] for i in 1:ncomparison])

        saveobj = Dict("delta" => delta,
                        "ARL0" => ARL0,
                        "par1SPSA" => opt_spsa[:opt][1], 
                        "nIterICspsa" => opt_spsa[:nIterIC],
                        "nIterOCspsa" => opt_spsa[:nIterOC],
                        "par1Bisec" => opt_bisection[:opt][1], 
                        "nIterICbisec" => opt_bisection[:nIterIC],
                        "nIterOCbisec" => opt_bisection[:nIterOC],
                        "RLspsa" => OC_rl_spsa,
                        "RLbisec" => OC_rl_bisec,
                        "time_spsa" => dt_spsa,
                        "time_bisec" => dt_bisec,
                        )
        safesave(filepath, saveobj)
    end

    filename = savename(Dict("delta" => delta, "ARL0" => ARL0, "SD" => SD, "type" => "NelderMead"), "jld2")
    filepath = datadir(simsDir_nelder, "EWMA", filename)
    if isfile(filepath)
        
    else
        #? First run for compilation
        NelderMeadOpt([0.2], minPar, maxPar, runEWMA, ARL0, shift=delta, maxiter=1, seed=seed)

        t_nelder = time()
        opt_nelder = NelderMeadOpt([0.2], minPar, maxPar, runEWMA, ARL0, shift=delta, seed=seed)
        dt_nelder = time() - t_nelder
        h_opt_nelder = saCL(1.0, (h,m,d) -> runEWMA(h,m,d, par=opt_nelder[2]), ARL0)[:h]

        seed = abs(rand(rng, Int))
        OC_rl_nelder = mean([runEWMA(h_opt_nelder, 10^5, 0.0, par=opt_nelder[2], shift=delta, seed = seed+i)[:rl] for i in 1:ncomparison])

        saveobj = Dict("delta" => delta,
                        "ARL0" => ARL0,
                        "par1Nelder" => opt_nelder[2][1],
                        "RLnelder" => OC_rl_nelder,
                        "time_nelder" => dt_nelder,
                        )
        safesave(filepath, saveobj)
    end
end

# ---- Bisection MEWMA2 ---- #
println("--- Starting MEWMA2 ---")
rng = MersenneTwister(SD)
deltas = [0.5, 1.0, 1.5, 2.0]
ARL0 = 370
nsim_bisection = 10000
minPar = fill(0.001, 2)
maxPar = fill(0.999, 2)
ncomparison = 10^6
for delta in deltas
    seed = abs(rand(rng, Int))
    filename = savename(Dict("delta" => delta, "ARL0" => ARL0, "SD" => SD), "jld2")
    filepath = datadir(simsDir, "MEWMA2", filename)
    if isfile(filepath)
        
    else
        #? First run for compilation
        bisectionOpt(minPar, maxPar, runMEWMA2, ARL0, shift = delta, maxiter=1, m = 3, verbose = false, nsim = 1, seed = seed, eps3 = abstol)


        t_bisec = time()
        opt_bisection = bisectionOpt(minPar, maxPar, runMEWMA2, ARL0, shift = delta, m = 3, verbose = false, nsim = nsim_bisection, seed = seed)
        dt_bisec = time() - t_bisec

        par0 = fill(0.2, 2)
        t1_spsa = time()
        optH = saCL(10.0, (h,m,d) -> runMEWMA2(h,m,d, par=par0), ARL0, verbose=false)
        dt1_spsa = time() - t1_spsa
        OPTS_run = OptimizationSettings(OPTS,
                    hStart = optH[:h],
                    initialStepSize = fill(0.1, 2),
                    minPar = minPar,
                    maxPar = maxPar,
                    comparison = nothing
                    )

        PROC = ProcessSettings(par0 = par0, ARL0 = ARL0, delta=delta, seed=seed)
        settings = Settings(OPTS_run, PROC)

        settings_compile = Settings(settings, OPTIM = OptimizationSettings(OPTS_run, maxiterSPSA=1))
        #? First run for compilation
        spsaCL(settings_compile, runMEWMA2)

        t_spsa = time()
        opt_spsa = spsaCL(settings, runMEWMA2)
        dt_spsa = time() - t_spsa
        dt_spsa += dt1_spsa


        h_opt_b = saCL(1.0, (h,m,d) -> runMEWMA2(h,m,d, par=opt_bisection[:opt]), ARL0)[:h]
        h_opt_spsa = saCL(1.0, (h,m,d) -> runMEWMA2(h,m,d, par=opt_spsa[:opt]), ARL0)[:h]

        seed = abs(rand(rng, Int))
        OC_rl_bisec = mean([runMEWMA2(h_opt_b, 10^5, 0.0, par=opt_bisection[:opt], shift=delta, seed = seed+i)[:rl] for i in 1:ncomparison])
        OC_rl_spsa = mean([runMEWMA2(h_opt_spsa, 10^5, 0.0, par=opt_spsa[:opt], shift=delta, seed = seed+i)[:rl] for i in 1:ncomparison])

        saveobj = Dict("delta" => delta,
                        "ARL0" => ARL0,
                        "par1SPSA" => opt_spsa[:opt][1], 
                        "par2SPSA" => opt_spsa[:opt][2], 
                        "nIterICspsa" => opt_spsa[:nIterIC],
                        "nIterOCspsa" => opt_spsa[:nIterOC],
                        "par1Bisec" => opt_bisection[:opt][1], 
                        "par2Bisec" => opt_bisection[:opt][2], 
                        "nIterICbisec" => opt_bisection[:nIterIC],
                        "nIterOCbisec" => opt_bisection[:nIterOC],
                        "RLspsa" => OC_rl_spsa,
                        "RLbisec" => OC_rl_bisec,
                        "time_spsa" => dt_spsa,
                        "time_bisec" => dt_bisec,
                        )
        safesave(filepath, saveobj)
    end

    filename = savename(Dict("delta" => delta, "ARL0" => ARL0, "SD" => SD, "type" => "NelderMead"), "jld2")
    filepath = datadir(simsDir_nelder, "MEWMA2", filename)
    if isfile(filepath)
        
    else
        #? First run for compilation
        NelderMeadOpt([0.2, 0.2], minPar, maxPar, runMEWMA2, ARL0, shift=delta, maxiter=1, seed=seed)

        t_nelder = time()
        opt_nelder = NelderMeadOpt([0.2, 0.2], minPar, maxPar, runMEWMA2, ARL0, shift=delta, seed=seed)
        dt_nelder = time() - t_nelder
        h_opt_nelder = saCL(1.0, (h,m,d) -> runMEWMA2(h,m,d, par=opt_nelder[2]), ARL0)[:h]

        seed = abs(rand(rng, Int))
        OC_rl_nelder = mean([runMEWMA2(h_opt_nelder, 10^5, 0.0, par=opt_nelder[2], shift=delta, seed = seed+i)[:rl] for i in 1:ncomparison])

        saveobj = Dict("delta" => delta,
                        "ARL0" => ARL0,
                        "par1Nelder" => opt_nelder[2][1],
                        "par2Nelder" => opt_nelder[2][2],
                        "RLnelder" => OC_rl_nelder,
                        "time_nelder" => dt_nelder,
                        )
        safesave(filepath, saveobj)
    end
end


# ---- Bisection MEWMA3 ---- #
println("--- Starting MEWMA3 ---")
rng = MersenneTwister(SD)
deltas = [0.5, 1.0, 1.5, 2.0]
ARL0 = 370
nsim_bisection = 10000
minPar = fill(0.001, 3)
maxPar = fill(0.999, 3)
ncomparison = 10^6
for delta in deltas
    seed = abs(rand(rng, Int))
    filename = savename(Dict("delta" => delta, "ARL0" => ARL0, "SD" => SD), "jld2")
    filepath = datadir(simsDir, "MEWMA3", filename)
    if isfile(filepath)
        
    else
        #? First run for compilation
        bisectionOpt(minPar, maxPar, runMEWMA3, ARL0, shift = delta, maxiter=1, m = 3, verbose = false, nsim = 1, seed = seed, eps3 = abstol)


        t_bisec = time()
        opt_bisection = bisectionOpt(minPar, maxPar, runMEWMA3, ARL0, shift = delta, m = 3, verbose = false, nsim = nsim_bisection, seed = seed)
        dt_bisec = time() - t_bisec

        par0 = fill(0.2, 3)
        t1_spsa = time()
        optH = saCL(10.0, (h,m,d) -> runMEWMA3(h,m,d, par=par0), ARL0, verbose=false)
        dt1_spsa = time() - t1_spsa
        OPTS_run = OptimizationSettings(OPTS,
                    hStart = optH[:h],
                    initialStepSize = fill(0.1, 3),
                    type = [:Numeric for _ in 1:3],
                    minPar = minPar,
                    maxPar = maxPar,
                    comparison = nothing
                    )

        PROC = ProcessSettings(par0 = par0, ARL0 = ARL0, delta=delta, seed=seed)
        settings = Settings(OPTS_run, PROC)

        settings_compile = Settings(settings, OPTIM = OptimizationSettings(OPTS_run, maxiterSPSA=1))
        #? First run for compilation
        spsaCL(settings_compile, runMEWMA3)

        t_spsa = time()
        opt_spsa = spsaCL(settings, runMEWMA3)
        dt_spsa = time() - t_spsa
        dt_spsa += dt1_spsa


        h_opt_b = saCL(1.0, (h,m,d) -> runMEWMA3(h,m,d, par=opt_bisection[:opt]), ARL0)[:h]
        h_opt_spsa = saCL(1.0, (h,m,d) -> runMEWMA3(h,m,d, par=opt_spsa[:opt]), ARL0)[:h]

        seed = abs(rand(rng, Int))
        OC_rl_bisec = mean([runMEWMA3(h_opt_b, 10^5, 0.0, par=opt_bisection[:opt], shift=delta, seed = seed+i)[:rl] for i in 1:ncomparison])
        OC_rl_spsa = mean([runMEWMA3(h_opt_spsa, 10^5, 0.0, par=opt_spsa[:opt], shift=delta, seed = seed+i)[:rl] for i in 1:ncomparison])

        saveobj = Dict("delta" => delta,
                        "ARL0" => ARL0,
                        "par1SPSA" => opt_spsa[:opt][1], 
                        "par2SPSA" => opt_spsa[:opt][2], 
                        "par3SPSA" => opt_spsa[:opt][3], 
                        "nIterICspsa" => opt_spsa[:nIterIC],
                        "nIterOCspsa" => opt_spsa[:nIterOC],
                        "par1Bisec" => opt_bisection[:opt][1], 
                        "par2Bisec" => opt_bisection[:opt][2], 
                        "par3Bisec" => opt_bisection[:opt][3], 
                        "nIterICbisec" => opt_bisection[:nIterIC],
                        "nIterOCbisec" => opt_bisection[:nIterOC],
                        "RLspsa" => OC_rl_spsa,
                        "RLbisec" => OC_rl_bisec,
                        "time_spsa" => dt_spsa,
                        "time_bisec" => dt_bisec,
                        )
        safesave(filepath, saveobj)
    end
    filename = savename(Dict("delta" => delta, "ARL0" => ARL0, "SD" => SD, "type" => "NelderMead"), "jld2")
    filepath = datadir(simsDir_nelder, "MEWMA3", filename)
    if isfile(filepath)
        
    else
        #? First run for compilation
        NelderMeadOpt([0.2, 0.2, 0.2], minPar, maxPar, runMEWMA3, ARL0, shift=delta, maxiter=1, seed = seed)

        t_nelder = time()
        opt_nelder = NelderMeadOpt([0.2, 0.2, 0.2], minPar, maxPar, runMEWMA3, ARL0, shift=delta, seed=seed)
        dt_nelder = time() - t_nelder
        h_opt_nelder = saCL(1.0, (h,m,d) -> runMEWMA3(h,m,d, par=opt_nelder[2]), ARL0)[:h]

        seed = abs(rand(rng, Int))
        OC_rl_nelder = mean([runMEWMA3(h_opt_nelder, 10^5, 0.0, par=opt_nelder[2], shift=delta, seed = seed+i)[:rl] for i in 1:ncomparison])

        saveobj = Dict("delta" => delta,
                        "ARL0" => ARL0,
                        "par1Nelder" => opt_nelder[2][1],
                        "par2Nelder" => opt_nelder[2][2],
                        "par3Nelder" => opt_nelder[2][3],
                        "RLnelder" => OC_rl_nelder,
                        "time_nelder" => dt_nelder,
                        )
        safesave(filepath, saveobj)
    end
end




println("--- Starting CUSUM with estimated parameter ---")
rng = MersenneTwister(SD)
deltas = 0.5:0.25:3.0
ARL0 = 370
nsim_bisection = 10000
minPar = [0.001]
maxPar = [4.0]
m0 = 500
for delta in deltas
    seed = abs(rand(rng, Int))
    filename = savename(Dict("delta" => delta, "ARL0" => ARL0, "SD" => SD), "jld2")
    filepath = datadir(simsDir, "CUSUM_est", filename)
    x = randn(m0)
    est = [mean(x), std(x)]
    runCUSUM_est = (h,m,d; par, x=x, shift=0.0, est=est, seed = Int(sample(1:1e06))) -> runCUSUM_boot(h,m,d, par=par, shift=shift, est=est, seed=seed, x=x)

    if isfile(filepath)
        
    else

        #? First run for compilation
        bisectionOpt(minPar, maxPar, runCUSUM_est, ARL0, shift = delta, maxiter=1, m = 3, verbose = false, nsim = 1, seed = seed, eps3 = abstol)

        t_bisec = time()
        opt_bisection = bisectionOpt(minPar, maxPar, runCUSUM_est, ARL0, shift = delta, m = 3, verbose = true, nsim = nsim_bisection, seed = seed)
        dt_bisec = time() - t_bisec


        par0 = [1.0]
        t1_spsa = time()
        optH = saCL(0.1, (h,m,d) -> runCUSUM_est(h,m,d, par=par0), ARL0, verbose=false)
        dt1_spsa = time() - t1_spsa
        OPTS_run = OptimizationSettings(OPTS,
                    hStart = optH[:h],
                    initialStepSize = [0.2],
                    minPar = minPar,
                    maxPar = maxPar,
                    #! Put comparison to compare accuracy
                    comparison = nothing
                    )

        PROC = ProcessSettings(par0 = par0, ARL0 = ARL0, delta=delta, seed=seed)
        settings = Settings(OPTS_run, PROC)

        settings_compile = Settings(settings, OPTIM = OptimizationSettings(OPTS_run, maxiterSPSA=1))

        #? First run for compilation
        spsaCL(settings_compile, runCUSUM_est)

        t_spsa = time()
        opt_spsa = spsaCL(settings, runCUSUM_est)
        dt_spsa = time() - t_spsa
        dt_spsa += dt1_spsa


        saveobj = Dict("delta" => delta,
                        "ARL0" => ARL0,
                        "parSPSA" => opt_spsa[:opt][1], 
                        "nIterICspsa" => opt_spsa[:nIterIC] + optH[:iter],
                        "nIterOCspsa" => opt_spsa[:nIterOC],
                        "parBisec" => opt_bisection[:opt][1], 
                        "nIterICbisec" => opt_bisection[:nIterIC],
                        "nIterOCbisec" => opt_bisection[:nIterOC],
                        "time_spsa" => dt_spsa,
                        "time_bisec" => dt_bisec,
                        )
        safesave(filepath, saveobj)
    end

    filename = savename(Dict("delta" => delta, "ARL0" => ARL0, "SD" => SD, "type" => "NelderMead"), "jld2")
    filepath = datadir(simsDir_nelder, "CUSUM_est", filename)
    if isfile(filepath)
        
    else
        #? First run for compilation
        NelderMeadOpt([1.0], minPar, maxPar, runCUSUM_est, ARL0, shift=delta, maxiter=1, seed=seed)

        t_nelder = time()
        opt_nelder = NelderMeadOpt([1.0], minPar, maxPar, runCUSUM_est, ARL0, shift=delta, seed=seed)
        dt_nelder = time() - t_nelder
        h_opt_nelder = saCL(1.0, (h,m,d) -> runCUSUM_est(h,m,d, par=opt_nelder[2]), ARL0)[:h]

        seed = abs(rand(rng, Int))
        OC_rl_nelder = mean([runCUSUM_est(h_opt_nelder, 10^5, 0.0, par=opt_nelder[2], shift=delta, seed = seed+i)[:rl] for i in 1:ncomparison])

        saveobj = Dict("delta" => delta,
                        "ARL0" => ARL0,
                        "parNelder" => opt_nelder[2][1],
                        "RLnelder" => OC_rl_nelder,
                        "time_nelder" => dt_nelder,
                        )
        safesave(filepath, saveobj)
    end
end

