using Random

function calculate_score(rl, target, constraint::Symbol)
    if constraint == :ARL
        return (rl - target) / target
    elseif constraint == :Median
        return -(float(rl <= target) - 0.5)
    else
        error("Undefined constraint type")
    end
end

"""
	saCL(hstart::Real, rlsim::Function, Arl0::Real; kw...)

Computes the control limit for a control chart such that it satisfies E[RL] = Arl0.
"""
function saCL(hstart::Real, rlsim::Function, Arl0::Real; Nfixed=500, Afixed=0.1, Amin=0.1, Amax=100, deltaSA=0.1, q=0.55, gamma=0.02, Nmin=1000, z = 3.0, Cmrl=10.0, maxiter = 4e05, adaptive=true, verbose=true, constraint = :ARL)
    v = (z/gamma)^2
    eps = 1e-06
    u = 0.0

    h = hstart
    hm = score = s2 = D = 0.0

    if verbose println("Running SA ...") end

    # Stage 1 - adaptive gain estimation
    rlPlus = 0.0
    rlMinus = 0.0
    scorePlus  = 0.0
    scoreMinus = 0.0
    D = 0.0
    if adaptive
        if verbose println("Running adaptive gain ...") end
        for i in 1:Nfixed
            runLength = rlsim(h, Cmrl * Arl0 * sqrt(i + Nfixed), deltaSA)
            rl = runLength[:rl]
            score = calculate_score(rl, Arl0, constraint)
            h = max(eps, h - Afixed * score / (i^q))

            rlPlus = runLength[:rlPlus]
            scorePlus = calculate_score(rlPlus, Arl0, constraint)
            rlMinus = runLength[:rlMinus]
            scoreMinus = calculate_score(rlMinus, Arl0, constraint)
            D += (scorePlus - scoreMinus)/i
        end
        D = 1.0 / (max(1.0/Amax, min(1.0/Amin, D/(2.0*deltaSA))))
        if verbose println("Estimated gain D = $(D)") end
    else
        D = Amin    
        if verbose println("Using fixed gain D = $(D)") end
    end


    # Stage 2 - Stochastic approximations
    if verbose println("Running optimization ...") end
    i = 0
    while i < (maxiter + Nmin)
        if i % floor(maxiter / 20) == 0 && verbose
            println("i: $(i)/$(Int(trunc(maxiter)) + Nmin)\th: $(h)\thm: $(hm)")
        end
        i += 1
        runLength = rlsim(h, Cmrl * Arl0 * sqrt(i + Nfixed), 0.0)
        rl = runLength[:rl]
        score = calculate_score(rl, Arl0, constraint)
        h = max(eps, h - D * score / (i^q))
        if i > Nmin
            Ndenom = (i - Nmin)
            hm = hm + (h - hm) / Ndenom 
            s2 = s2 + (score * score - s2) / Ndenom
        end
        if (i > Nmin) && (i > v * maximum(s2))
            if verbose println("Convergence!") end
            break
        end
    end
    
    if verbose println("Done.") end

    return (h=hm, iter=i)
end
