using Random
using Distributions
using LinearAlgebra

function runCUSUM(h, maxrl = 100000, deltaSA=0.0; par = 2.0, shift = 0.0, est::Union{Nothing, Vector{Float64}} = nothing, seed=Int(sample(1:1e06)))
    Random.seed!(seed)
    i = 0
    R = 0.0; Rp = 0.0; Rm = 0.0;
    p = length(h)

    rl = rlPlus = rlMinus = maxrl
    notDoneRl = true
    notDoneP = notDoneM = (deltaSA > 0.0)
    notDone = notDoneRl + notDoneM + notDoneP
    k = first(par)
    while i < maxrl && (notDone > 0)
        i = i+1
        y = randn()
        if isnothing(est)
            R = max(0.0, R + y - k)
        else
            R = max(0.0, R + (y - est[1])/(est[2]) - k)
        end
        if notDoneRl && R > h
            notDoneRl = false
            rl = i
        end
        if notDoneP && R > h + deltaSA
            notDoneP = false
            rlPlus = i
        end
        if notDoneM && R > h - deltaSA
            notDoneM = false
            rlMinus = i
        end
        notDone = notDoneRl + notDoneM + notDoneP
    end

    return (rl = rl, rlPlus = rlPlus, rlMinus = rlMinus)
end
export runCUSUM


function runCUSUM_boot(h, maxrl = 100000, deltaSA=0.0; par = 2.0, shift = 0.0, x::Vector{Float64}, est::Vector{Float64} = nothing, seed=Int(sample(1:1e06)))
    Random.seed!(seed)
    i = 0
    R = 0.0; Rp = 0.0; Rm = 0.0;
    p = length(h)

    rl = rlPlus = rlMinus = maxrl
    notDoneRl = true
    notDoneP = notDoneM = (deltaSA > 0.0)
    notDone = notDoneRl + notDoneM + notDoneP
    k = first(par)
    idx = 0
    n = length(x)
    while i < maxrl && (notDone > 0)
        i = i+1
        idx = sample(1:n)
        y = x[idx] + shift
        if isnothing(est)
            R = max(0.0, R + y - k)
        else
            R = max(0.0, R + (y - est[1])/(est[2]) - k)
        end
        if notDoneRl && R > h
            notDoneRl = false
            rl = i
        end
        if notDoneP && R > h + deltaSA
            notDoneP = false
            rlPlus = i
        end
        if notDoneM && R > h - deltaSA
            notDoneM = false
            rlMinus = i
        end
        notDone = notDoneRl + notDoneM + notDoneP
    end

    return (rl = rl, rlPlus = rlPlus, rlMinus = rlMinus)
end
export runCUSUM_boot

function runEWMA(h, maxrl = 100000, deltaSA=0.0; par = 0.15, shift = 0.0, seed=Int(sample(1:1e06)))
    Random.seed!(seed)
    i = 0
    R = 0.0
    p = length(h)

    rl = rlPlus = rlMinus = maxrl
    notDoneRl = true
    notDoneP = notDoneM = (deltaSA > 0.0)
    notDone = notDoneRl + notDoneM + notDoneP
    λ = first(par)
    while i < maxrl && (notDone > 0)
        i = i+1
        y = randn() + first(shift)
        R = (1-λ)*R + λ*y
        if notDoneRl && abs(R) > h
            notDoneRl = false
            rl = i
        end
        if notDoneP && abs(R) > h+deltaSA
            notDoneP = false
            rlPlus = i
        end
        if notDoneM && abs(R) > h-deltaSA
            notDoneM = false
            rlMinus = i
        end
        notDone = notDoneRl + notDoneM + notDoneP
    end

    if deltaSA == 0.0
        rlPlus = rl
        rlMinus = rl
    end
    return (rl = rl, rlPlus = rlPlus, rlMinus = rlMinus)
end
export runEWMA


function runMEWMA3(h, maxrl = 100000, deltaSA=0.0; par = fill(0.1, 3), shift = 0.0, seed=Int(sample(1:1e06)))
    Random.seed!(seed)
    i = 0
    p = length(h)

    rl = rlPlus = rlMinus = maxrl
    notDoneRl = true
    notDoneP = notDoneM = (deltaSA > 0.0)
    notDone = notDoneRl + notDoneM + notDoneP
    R = zeros(3)
    mus = [0.0, 1.0, 1.0]
    dists = [Normal(mus[1],1), Chisq(mus[2]), Poisson(mus[3])]
    Sigma = float.([1 0 0; 0 1 0; 0 0 1])
    SigmaZ = similar(Sigma)
    for k in 1:3
        for l in 1:3
            #? Using asymptotic matrix (constant limit)
            SigmaZ[k,l] = par[k]*par[l]/(par[k] + par[l] - par[k]*par[l]) * Sigma[k,l]
        end
    end
    D1 = Normal(mus[1] + shift, 1)
    D2 = Chisq(mus[2] + shift*sqrt(2*mus[2]))
    D3 = Poisson(mus[3] + shift*sqrt(mus[3]))
    SigmaZInv = inv(SigmaZ)
    y = Vector{Float64}(undef, 3)
    T = 0.0
    while i < maxrl && (notDone > 0)
        T = 0.0
        i = i+1
        y[1] = rand(D1)
        y[2] = rand(D2)
        y[3] = rand(D3)

        for j in 1:3
            R[j] = (1.0 .- par[j]) .* R[j] + par[j] .* (y[j] - mus[j])
            T += R[j]^2 * SigmaZInv[j,j]
        end
        if notDoneRl && T > h
            notDoneRl = false
            rl = i
        end
        if notDoneP && T > (h+deltaSA)
            notDoneP = false
            rlPlus = i
        end
        if notDoneM && T > (h-deltaSA)
            notDoneM = false
            rlMinus = i
        end
        notDone = notDoneRl + notDoneM + notDoneP
    end

    if deltaSA == 0.0
        rlPlus = rl
        rlMinus = rl
    end
    return (rl = rl, rlPlus = rlPlus, rlMinus = rlMinus)
end
export runMEWMA3


function runMEWMA2(h, maxrl = 100000, deltaSA=0.0; par = fill(0.1, 2), shift = 0.0, seed=Int(sample(1:1e06)))
    Random.seed!(seed)
    i = 0
    p = length(h)

    rl = rlPlus = rlMinus = maxrl
    notDoneRl = true
    notDoneP = notDoneM = (deltaSA > 0.0)
    notDone = notDoneRl + notDoneM + notDoneP
    R = zeros(2)
    mus = [0.0, 1.0]
    dists = [Normal(mus[1],1), Chisq(mus[2])]
    Sigma = float.([1 0; 0 1;])
    SigmaZ = similar(Sigma)
    for k in 1:2
        for l in 1:2
            #? Using asymptotic matrix (constant limit)
            SigmaZ[k,l] = par[k]*par[l]/(par[k] + par[l] - par[k]*par[l]) * Sigma[k,l]
        end
    end
    D1 = Normal(mus[1] + shift, 1)
    D2 = Chisq(mus[2] + shift*sqrt(2*mus[2]))
    SigmaZInv = inv(SigmaZ)
    y = Vector{Float64}(undef, 2)
    T = 0.0
    while i < maxrl && (notDone > 0)
        T = 0.0
        i = i+1
        y[1] = rand(D1)
        y[2] = rand(D2)

        for j in 1:2
            R[j] = (1.0 .- par[j]) .* R[j] + par[j] .* (y[j] - mus[j])
            T += R[j]^2 * SigmaZInv[j,j]
        end
        if notDoneRl && T > h
            notDoneRl = false
            rl = i
        end
        if notDoneP && T > (h+deltaSA)
            notDoneP = false
            rlPlus = i
        end
        if notDoneM && T > (h-deltaSA)
            notDoneM = false
            rlMinus = i
        end
        notDone = notDoneRl + notDoneM + notDoneP
    end

    if deltaSA == 0.0
        rlPlus = rl
        rlMinus = rl
    end
    return (rl = rl, rlPlus = rlPlus, rlMinus = rlMinus)
end
export runMEWMA2


function augmented_vector(y, obs, mu, dist)
    notObs = setdiff(1:length(y), obs)
    A = sum(pdf.(dist, y[obs] .- mu) ./ pdf.(dist, y[obs]))
    xi, i = findmax(y)
    B = cdf(dist, xi)
    C = cdf(dist, xi - mu)
    d = length(notObs)
    eta = zeros(length(y))
    eta[i] = (B^d * A + B^(d-1) * C * d) / (A + d)
    eta[notObs] .= (1.0 - eta[i]) / d
    return eta
end
export augmented_vector

function runRSADA(h, maxrl = 100000, deltaSA=0.0; p = 100, par = [1.0, 1.5], shift = 0.0, tau = 0, q = 20, n = 10, whichOC = sample(1:p, n), seed=Int(sample(1:1e06)))
    Random.seed!(seed)
    i = 0
    R = 0.0; Rp = 0.0; Rm = 0.0;
    rVec = zeros(Int(ceil(maxrl)))
    k = par[1]
    mu_min = par[2]
    dist = Normal(0,1)
    rl = rlPlus = rlMinus = maxrl
    notDoneRl = true
    notDoneP = notDoneM = (deltaSA > 0.0)
    notDone = notDoneRl + notDoneM + notDoneP
    obs = sample(1:p, q)
    C = 0.0
    g = fill(1.0/p, p)
    S1 = zeros(p)
    S2 = zeros(p)
    while i < maxrl && (notDone > 0)
        i = i+1
        y = rand(dist, p)
        if i > tau
            y[whichOC] .+= shift
        end
        eta = augmented_vector(y, obs, mu_min, dist)
        S1 = S1 + eta
        S2 = S2 + g
        C = sum((S1 - S2).^2 ./ S2)
        if C <= k
            S1 = deepcopy(g)
            S2 = deepcopy(g)
        else
            C = (C-k)/C
            S1 = C * S1
            S2 = C * S2
        end
        obs = sortperm(-S1)[1:q]
        R = sum((S1 - S2).^2 ./ S2)
        rVec[i] = R


        if notDoneRl && R > h
            notDoneRl = false
            rl = i
        end
        if notDoneP && R > h + deltaSA
            notDoneP = false
            rlPlus = i
        end
        if notDoneM && R > h - deltaSA
            notDoneM = false
            rlMinus = i
        end
        notDone = notDoneRl + notDoneM + notDoneP
    end

    return (rl = rl, rlPlus = rlPlus, rlMinus = rlMinus)
end
export runRSADA
