using Distributions
using LinearAlgebra
using Random
using StatsBase

function discreteProject(par::Vector, type::Vector{Symbol})
    for j in 1:length(par)
        if type[j] == :Integer
            par[j] = round(par[j])
        end
    end
    return par
end

function objectiveFunction(par, hStart, delta, rlsim, ARL0, NfixedSA, NminSA, maxiterSA, gavg, seedSP; BIGINT = 100000, wts = fill(1.0, length(delta)), objective = :ARL, constraint = :ARL)
    if length(wts) != length(delta)
        error("Length mismatch between delta (length: $(length(delta))) and weights (length: $(length(weights))) ")
    end
    opt = saCL(hStart, (h, maxrl, deltaSA) -> rlsim(h, maxrl, deltaSA; par=par), ARL0, Nfixed = NfixedSA, Nmin = NminSA, maxiter = maxiterSA, verbose = false, constraint=constraint)
    hOpt = opt[:h]
    # @show hOpt
    # @show par
    nIterIC = opt[:iter]
    #? Using same seed to reduce random error noise
    rl = zeros(gavg)
    wts = wts ./ sum(wts)           # Normalize weights
    for i in 1:gavg
        for j in eachindex(delta)
            rl[i] += wts[j]*rlsim(hOpt, BIGINT, 0.0; par=par, shift=delta[j], seed=seedSP+i)[:rl]
        end
    end
    obj = 0.0
    if objective == :ARL
        obj = mean(rl)
    elseif objective == :Median
        obj = median(rl)
    else
        error("Unknown objective")
    end
    return (obj = obj, sd = std(rl)/sqrt(gavg), rl = rl, h = hOpt, nIterIC = nIterIC, nIterOC = length(delta)*gavg)
end
export objectiveFunction

"""
    spsaCL(par::Vector, rlsim::Function, Arl0::Real; kw...)

Computes the optimal parameter for a control chart against a shift of size delta.
`rlsim` has signature
    rlsim(h, maxrl, deltaSA; par)
"""
function spsaCL(par::Vector, rlsim::Function, ARL0::Real, delta; type::Vector{Symbol} = [:Numeric for _ in 1:length(par)], hStart = 1e-05, minPar = 0.005, maxPar = Inf, initialStepSize = 0.05, expNumberLossEval = 100, nAdaptiveGain = 20, Cmrl=10.0, maxiterSPSA = Integer(5e03), NfixedSA = 300, NminSA = 100, maxiterSA = 500, NfixedSPSA = 50, NminSPSA = 100, z = 3.0, gg = 0.02, gavg=500, abstol=1e-03, seed=12345, verbose=true, print_every=20, comparison::Union{Nothing, Int} = nothing, objective=:ARL, constraint=:ARL)

    parCurrent = deepcopy(par)
    parPlus = deepcopy(par)
    parMinus = deepcopy(par)
    bern = Bernoulli(0.5)
    p = length(par)
    perturb = Vector{Float64}(undef, p)

    eps = 0.005          # Minimum value for the parameter
    alpha = 0.602     # Slower decay
    gamma = 0.101     # Slower decay
    # alpha = 1.0         # Faster decay
    # gamma = 1/6         # Faster decay
    score = 0.0
    nIterIC = nIterOC = 0

    Random.seed!(seed)
    if verbose println("- Running SPSA ...") end
    wts = fill(1.0/length(delta), length(delta))

    # ----------------------------------------- #
    #        AUTOMATIC GAIN SELECTION           #
    # ----------------------------------------- #
    # Following the recommendation of https://www.jhuapl.edu/spsa/Pages/MATLAB.htm
    # The `step` variable is a crucial parameter not to jump out of the boundary right away

    if verbose println("--- Running automatic gain selection...") end
    
    A = 0.1*expNumberLossEval/2.0

    seedSP = Int(sample(1:1e06))
    c = objectiveFunction(parCurrent, hStart, delta, rlsim, ARL0, NfixedSA, NminSA, maxiterSA, gavg, seedSP, wts=wts, objective=objective, constraint = constraint)[:sd]
    # @show c
    c = min(c, 0.1)
    c = max(c, 0.001)
    c = min(c, minimum(abs.(parCurrent .- minPar)), minimum(abs.(parCurrent .- maxPar)))
    # @show c
    # c = 0.1      # Standard deviation of measurement noise
    
    gbar = zeros(length(par))
    rlPlus = rlMinus = 0.0
    for i in 1:nAdaptiveGain
        perturb = 2.0 * rand(bern, p) .- 1.0

        parPlus  = parCurrent .+ c*perturb
        parPlus = max.(minPar, parPlus)
        parPlus = min.(maxPar, parPlus)
        parPlus = discreteProject(parPlus, type)
        
        parMinus = parCurrent .- c*perturb
        parMinus = max.(minPar, parMinus)
        parMinus = min.(maxPar, parMinus)
        parMinus = discreteProject(parMinus, type)
        # @show parPlus, parMinus
        
        seedSP = Int(sample(1:1e06))
        rlPlus = objectiveFunction(parPlus, hStart, delta, rlsim, ARL0, NfixedSA, NminSA, maxiterSA, gavg, seedSP, wts=wts, objective=objective, constraint = constraint)
        nIterIC += rlPlus[:nIterIC]
        nIterOC += rlPlus[:nIterOC]
        rlMinus = objectiveFunction(parMinus, hStart, delta, rlsim, ARL0, NfixedSA, NminSA, maxiterSA, gavg, seedSP, wts=wts, objective=objective, constraint = constraint)
        nIterIC += rlMinus[:nIterIC]
        nIterOC += rlMinus[:nIterOC]
        score = (rlPlus[:obj] - rlMinus[:obj]) ./ (2.0 * c * perturb)
        gbar = gbar + abs.(score)
    end
    meangbar = mean(gbar) / (nAdaptiveGain)
    # Avoid non-convergence when meangbar is estimated too small
    if meangbar == 0.0
       meangbar = 0.1 
    end
    a = initialStepSize .* ((A + 1)^alpha)/meangbar

    if verbose
        println("a: $(round.(a, digits=5))\tA: $(A)\tc: $(c)\th: $(hStart)")
        println("--- ... Completed.")
    end

    s2 = zeros(p)
    VV = zeros(p); parPolyakRuppert = zeros(p); oldPolyakRuppert = zeros(p);
    parFinal = zeros(p)
    v = (z/gg)^2

    convergence = "None"
    if verbose println("--- Running optimization routine...") end
    i = 1
    if verbose
        println("iter: $(i)/$(maxiterSPSA),\tCurrent: $(round.(parCurrent, digits=5))\tPolyak average: $(round.(parPolyakRuppert, digits=5))\tCriterion: $(NfixedSPSA + v*(maximum(s2)))")
    end
    while i < maxiterSPSA
        if i % print_every == 0 && verbose
            println("iter: $(i)/$(maxiterSPSA),\tCurrent: $(round.(parCurrent, digits=5))\tPolyak average: $(round.(parPolyakRuppert, digits=5))\tCriterion: $(NfixedSPSA + v*(maximum(s2)))")
        end

        c_k = c/(i^gamma)
        perturb = 2.0 * rand(bern, p) .- 1.0
        parPlus  = parCurrent .+ c_k*perturb
        parPlus = max.(minPar, parPlus)
        parPlus = min.(maxPar, parPlus)
        parPlus = discreteProject(parPlus, type)

        parMinus = max.(eps, parCurrent .- c_k*perturb)
        parMinus = max.(minPar, parMinus)
        parMinus = min.(maxPar, parMinus)
        parMinus = discreteProject(parMinus, type)

        # @show hStart
        # @show parPlus, parMinus
        # @show hOptPlus, hOptMinus

        seedSP = Int(sample(1:1e06))
        rlPlus = objectiveFunction(parPlus, hStart, delta, rlsim, ARL0, NfixedSA, NminSA, maxiterSA, gavg, seedSP, wts=wts, objective=objective, constraint = constraint)
        nIterIC += rlPlus[:nIterIC]
        nIterOC += rlPlus[:nIterOC]
        rlMinus = objectiveFunction(parMinus, hStart, delta, rlsim, ARL0, NfixedSA, NminSA, maxiterSA, gavg, seedSP, wts=wts, objective=objective, constraint = constraint)
        nIterIC += rlMinus[:nIterIC]
        nIterOC += rlMinus[:nIterOC]

        EA = mean((rlPlus[:rl] - rlMinus[:rl]) ./ (2.0 * c_k))
        score = EA *  perturb
        # if i % print_every == 0 && verbose
        #     println("\trlPlus: $(rlPlus)\trlMinus:$(rlMinus)\tScore: $(score)")
        # end
        # @show rlPlus, rlMinus
        # @show score
        # println("")
        a_k = a ./ (i+A)^alpha
        parCurrent = parCurrent - a_k .* score
        parCurrent = max.(minPar, parCurrent)
        parCurrent = min.(maxPar, parCurrent)
        parCurrent = discreteProject(parCurrent, type)

        VA = var(((rlPlus[:rl] - rlMinus[:rl]) ./ (2.0 * c_k)))/gavg
        VV = score.^2


        if i > NfixedSPSA
            copyto!(oldPolyakRuppert, parPolyakRuppert)
            Ndenom = (i - NfixedSPSA)
            parPolyakRuppert .+= (parCurrent .- parPolyakRuppert) / Ndenom 
            s2 += (VV .- s2) / Ndenom
            # @show score, VV, s2
        end
        hStart = min(rlPlus[:h], rlMinus[:h])

        if isnothing(comparison)
            if i > NminSPSA && norm(parPolyakRuppert - oldPolyakRuppert) < abstol
                convergence = "Absolute tolerance criterion."
                parFinal = parPolyakRuppert
                break
            end
            if i > NminSPSA && i > v*(maximum(s2))
                convergence = "Stochastic tolerance criterion."
                parFinal = parPolyakRuppert
                break
            end
        else
            if nIterIC + nIterOC >= comparison
                convergence = "Convergence for comparison"
                parFinal = parPolyakRuppert
                break
            end
        end
        i += 1
    end
    
    if verbose println("--- ... Completed.") end

    return (opt = parFinal, iter=i, convergence=convergence, nIterIC = nIterIC, nIterOC = nIterOC)
end
